# coding: utf-8
"""
Tokenizer module
"""
import logging
import shutil
from pathlib import Path
from typing import Dict, List

import sentencepiece as sp
from subword_nmt import apply_bpe

from joeynmt.constants import BOS_TOKEN, EOS_TOKEN, PAD_TOKEN, UNK_TOKEN
from joeynmt.helpers import ConfigurationError, remove_extra_spaces, unicode_normalize

logger = logging.getLogger(__name__)


class BasicTokenizer:
    SPACE = chr(32)  # ' ': half-width white space (ascii)
    SPACE_ESCAPE = chr(9601)  # '▁': sentencepiece default
    SPECIALS = [BOS_TOKEN, EOS_TOKEN, PAD_TOKEN, UNK_TOKEN]

    def __init__(
        self,
        level: str = "word",
        lowercase: bool = False,
        normalize: bool = False,
        max_length: int = -1,
        min_length: int = -1,
        **kwargs,
    ):
        # pylint: disable=unused-argument
        self.level = level
        self.lowercase = lowercase
        self.normalize = normalize

        # filter by length
        self.max_length = max_length
        self.min_length = min_length

        # pretokenizer
        self.pretokenizer = kwargs.get("pretokenizer", "none").lower()
        assert self.pretokenizer in ["none", "moses"], \
            "Currently, we support moses tokenizer only."
        # sacremoses
        if self.pretokenizer == "moses":
            try:
                from sacremoses import (  # pylint: disable=import-outside-toplevel
                    MosesTokenizer, MosesDetokenizer, MosesPunctNormalizer,
                )
            except ImportError as e:
                logger.error(e)
                raise ImportError from e

            self.lang = kwargs.get("lang", "en")
            self.moses_tokenizer = MosesTokenizer(lang=self.lang)
            self.moses_detokenizer = MosesDetokenizer(lang=self.lang)
            if self.normalize:
                self.moses_normalizer = MosesPunctNormalizer()

    def pre_process(self, raw_input: str) -> str:
        """
        Pre-process text
            - ex.) Lowercase, Normalize, Remove emojis,
                Pre-tokenize(add extra white space before punc) etc.
            - applied for all inputs both in training and inference.
        """
        if self.normalize:
            raw_input = remove_extra_spaces(unicode_normalize(raw_input))

        if self.pretokenizer == "moses":
            if self.normalize:
                raw_input = self.moses_normalizer.normalize(raw_input)
            raw_input = self.moses_tokenizer.tokenize(raw_input, return_str=True)

        if self.lowercase:
            raw_input = raw_input.lower()

        # ensure the string is not empty.
        assert raw_input is not None and len(raw_input) > 0, raw_input
        return raw_input

    def __call__(self, raw_input: str, is_train: bool = False) -> List[str]:
        """Tokenize single sentence"""
        if self.level == "word":
            sequence = raw_input.split(self.SPACE)
        elif self.level == "char":
            sequence = list(raw_input.replace(self.SPACE, self.SPACE_ESCAPE))

        if is_train and self._filter_by_length(len(sequence)):
            return None
        return sequence

    def _filter_by_length(self, length: int) -> bool:
        """
        Check if the given seq length is out of the valid range.

        :param length: (int) number of tokens
        :return: True if the length is invalid(= to be filtered out), False if valid.
        """
        return length > self.max_length > 0 or self.min_length > length > 0

    def _remove_special(self, sequence: List[str], generate_unk: bool = False):
        specials = self.SPECIALS[:-1] if generate_unk else self.SPECIALS
        return [token for token in sequence if token not in specials]

    def post_process(self, sequence: List[str], generate_unk: bool = True) -> str:
        """Detokenize"""
        sequence = self._remove_special(sequence, generate_unk=generate_unk)
        if self.level == "word":
            if self.pretokenizer == "moses":
                detokenized = self.moses_detokenizer.detokenize(sequence)
            else:
                detokenized = self.SPACE.join(sequence)
        elif self.level == "char":
            detokenized = "".join(sequence).replace(self.SPACE_ESCAPE, self.SPACE)

        # Remove extra spaces
        if self.normalize:
            detokenized = remove_extra_spaces(detokenized)

        # ensure the string is not empty.
        assert detokenized is not None and len(detokenized) > 0, detokenized
        return detokenized

    def set_vocab(self, itos: List[str]) -> None:
        """
        Set vocab
        :param itos: (list) indices-to-symbols mapping
        """
        pass  # pylint: disable=unnecessary-pass

    def __repr__(self):
        return (f"{self.__class__.__name__}(level={self.level}, "
                f"lowercase={self.lowercase}, normalize={self.normalize}, "
                f"filter_by_length=({self.min_length}, {self.max_length}), "
                f"pretokenizer={self.pretokenizer})")


class SentencePieceTokenizer(BasicTokenizer):

    def __init__(
        self,
        level: str = "bpe",
        lowercase: bool = False,
        normalize: bool = False,
        max_length: int = -1,
        min_length: int = -1,
        **kwargs,
    ):
        super().__init__(level, lowercase, normalize, max_length, min_length, **kwargs)
        assert self.level == "bpe"

        self.model_file: Path = Path(kwargs["model_file"])
        assert self.model_file.is_file(), f"model file {self.model_file} not found."

        self.spm = sp.SentencePieceProcessor()
        self.spm.load(kwargs["model_file"])

        self.nbest_size: int = kwargs.get("nbest_size", 5)
        self.alpha: float = kwargs.get("alpha", 0.0)

    def __call__(self, raw_input: str, is_train: bool = False) -> List[str]:
        """Tokenize"""
        if is_train and self.alpha > 0:
            tokenized = self.spm.sample_encode_as_pieces(
                raw_input,
                nbest_size=self.nbest_size,
                alpha=self.alpha,
            )
        else:
            tokenized = self.spm.encode(raw_input, out_type=str)

        if is_train and self._filter_by_length(len(tokenized)):
            return None
        return tokenized

    def post_process(self, sequence: List[str], generate_unk: bool = True) -> str:
        """Detokenize"""
        sequence = self._remove_special(sequence, generate_unk=generate_unk)

        # Decode back to str
        detokenized = self.spm.decode(sequence)
        detokenized = detokenized.replace(self.SPACE_ESCAPE, self.SPACE).strip()

        # Apply moses detokenizer
        if self.pretokenizer == "moses":
            detokenized = self.moses_detokenizer.detokenize(detokenized.split())

        # Remove extra spaces
        if self.normalize:
            detokenized = remove_extra_spaces(detokenized)

        # ensure the string is not empty.
        assert detokenized is not None and len(detokenized) > 0, detokenized
        return detokenized

    def set_vocab(self, itos: List[str]) -> None:
        """Set vocab"""
        self.spm.SetVocabulary(itos)

    def copy_cfg_file(self, model_dir: Path) -> None:
        """Copy confg file to model_dir"""
        if (model_dir / self.model_file.name).is_file():
            logger.warning(
                "%s already exists. Stop copying.",
                (model_dir / self.model_file.name).as_posix(),
            )
        shutil.copy2(self.model_file, (model_dir / self.model_file.name).as_posix())

    def __repr__(self):
        return (f"{self.__class__.__name__}(level={self.level}, "
                f"lowercase={self.lowercase}, normalize={self.normalize}, "
                f"filter_by_length=({self.min_length}, {self.max_length}), "
                f"pretokenizer={self.pretokenizer}, "
                f"tokenizer={self.spm.__class__.__name__}, "
                f"nbest_size={self.nbest_size}, alpha={self.alpha})")


class SubwordNMTTokenizer(BasicTokenizer):

    def __init__(
        self,
        level: str = "bpe",
        lowercase: bool = False,
        normalize: bool = False,
        max_length: int = -1,
        min_length: int = -1,
        **kwargs,
    ):
        super().__init__(level, lowercase, normalize, max_length, min_length, **kwargs)
        assert self.level == "bpe"

        self.codes: Path = Path(kwargs["codes"])
        assert self.codes.is_file(), f"codes file {self.codes} not found."

        self.separator: str = kwargs.get("separator", "@@")
        bpe_parser = apply_bpe.create_parser()
        bpe_args = bpe_parser.parse_args(
            ["--codes", kwargs["codes"], "--separator", self.separator])
        self.bpe = apply_bpe.BPE(
            bpe_args.codes,
            bpe_args.merges,
            bpe_args.separator,
            None,
            bpe_args.glossaries,
        )
        self.dropout: float = kwargs.get("dropout", 0.0)

    def __call__(self, raw_input: str, is_train: bool = False) -> List[str]:
        """Tokenize"""
        dropout = self.dropout if is_train else 0.0
        tokenized = self.bpe.process_line(raw_input, dropout).strip().split()
        if is_train and self._filter_by_length(len(tokenized)):
            return None
        return tokenized

    def post_process(self, sequence: List[str], generate_unk: bool = True) -> str:
        """Detokenize"""
        sequence = self._remove_special(sequence, generate_unk=generate_unk)

        # Remove separators, join with spaces
        detokenized = self.SPACE.join(sequence).replace(self.separator + self.SPACE, "")
        # Remove final merge marker.
        if detokenized.endswith(self.separator):
            detokenized = detokenized[:-2]

        # Moses detokenizer
        if self.pretokenizer == "moses":
            detokenized = self.moses_detokenizer.detokenize(detokenized.split())

        # Remove extra spaces
        if self.normalize:
            detokenized = remove_extra_spaces(detokenized)

        # ensure the string is not empty.
        assert detokenized is not None and len(detokenized) > 0, detokenized
        return detokenized

    def set_vocab(self, itos: List[str]) -> None:
        """Set vocab"""
        vocab = set(itos) - set(self.SPECIALS)
        self.bpe.vocab = vocab

    def copy_cfg_file(self, model_dir: Path) -> None:
        """Copy confg file to model_dir"""
        shutil.copy2(self.codes, (model_dir / self.codes.name).as_posix())

    def __repr__(self):
        return (f"{self.__class__.__name__}(level={self.level}, "
                f"lowercase={self.lowercase}, normalize={self.normalize}, "
                f"filter_by_length=({self.min_length}, {self.max_length}), "
                f"pretokenizer={self.pretokenizer}, "
                f"tokenizer={self.bpe.__class__.__name__}, "
                f"separator={self.separator}, dropout={self.dropout})")


def _build_tokenizer(cfg: Dict) -> BasicTokenizer:
    """Builds tokenizer."""
    tokenizer = None
    tokenizer_cfg = cfg.get("tokenizer_cfg", {})

    # assign lang for moses tokenizer
    if tokenizer_cfg.get("pretokenizer", "none") == "moses":
        tokenizer_cfg["lang"] = cfg["lang"]

    if cfg["level"] in ["word", "char"]:
        tokenizer = BasicTokenizer(
            level=cfg["level"],
            lowercase=cfg.get("lowercase", False),
            normalize=cfg.get("normalize", False),
            max_length=cfg.get("max_length", -1),
            min_length=cfg.get("min_length", -1),
            **tokenizer_cfg,
        )
    elif cfg["level"] == "bpe":
        tokenizer_type = cfg.get("tokenizer_type", cfg.get("bpe_type", "sentencepiece"))
        if tokenizer_type == "sentencepiece":
            assert "model_file" in tokenizer_cfg
            tokenizer = SentencePieceTokenizer(
                level=cfg["level"],
                lowercase=cfg.get("lowercase", False),
                normalize=cfg.get("normalize", False),
                max_length=cfg.get("max_length", -1),
                min_length=cfg.get("min_length", -1),
                **tokenizer_cfg,
            )
        elif tokenizer_type == "subword-nmt":
            assert "codes" in tokenizer_cfg
            tokenizer = SubwordNMTTokenizer(
                level=cfg["level"],
                lowercase=cfg.get("lowercase", False),
                normalize=cfg.get("normalize", False),
                max_length=cfg.get("max_length", -1),
                min_length=cfg.get("min_length", -1),
                **tokenizer_cfg,
            )
        else:
            raise ConfigurationError(f"{tokenizer_type}: Unknown tokenizer type.")
    else:
        raise ConfigurationError(f"{cfg['level']}: Unknown tokenization level.")
    return tokenizer


def build_tokenizer(data_cfg: Dict) -> Dict[str, BasicTokenizer]:
    src_lang = data_cfg["src"]["lang"]
    trg_lang = data_cfg["trg"]["lang"]
    tokenizer = {
        src_lang: _build_tokenizer(data_cfg["src"]),
        trg_lang: _build_tokenizer(data_cfg["trg"]),
    }
    logger.info("%s tokenizer: %s", src_lang, tokenizer[src_lang])
    logger.info("%s tokenizer: %s", trg_lang, tokenizer[trg_lang])
    return tokenizer



# coding: utf-8
"""
Data module
"""
import logging
from functools import partial
from typing import Callable, List, Optional, Tuple

import torch
from torch.utils.data import (
    BatchSampler,
    DataLoader,
    Dataset,
    RandomSampler,
    Sampler,
    SequentialSampler,
)

from joeynmt.batch import Batch
from joeynmt.constants import PAD_ID
from joeynmt.datasets import build_dataset
from joeynmt.helpers import log_data_info
from joeynmt.tokenizers import build_tokenizer
from joeynmt.vocabulary import Vocabulary, build_vocab

logger = logging.getLogger(__name__)
CPU_DEVICE = torch.device("cpu")


def load_data(
    data_cfg: dict,
    datasets: list = None
) -> Tuple[Vocabulary, Vocabulary, Optional[Dataset], Optional[Dataset],
           Optional[Dataset], ]:
    """
    Load train, dev and optionally test data as specified in configuration.
    Vocabularies are created from the training set with a limit of `voc_limit` tokens
    and a minimum token frequency of `voc_min_freq` (specified in the configuration
    dictionary).

    The training data is filtered to include sentences up to `max_sent_length` on source
    and target side.

    If you set `random_{train|dev}_subset`, a random selection of this size is used
    from the {train|development} set instead of the full {train|development} set.

    :param data_cfg: configuration dictionary for data ("data" part of config file)
    :param datasets: list of dataset names to load
    :returns:
        - src_vocab: source vocabulary
        - trg_vocab: target vocabulary
        - train_data: training dataset
        - dev_data: development dataset
        - test_data: test dataset if given, otherwise None
    """
    if datasets is None:
        datasets = ["train", "dev", "test"]
    src_cfg = data_cfg["src"]
    trg_cfg = data_cfg["trg"]

    # load data from files
    src_lang = src_cfg["lang"]
    trg_lang = trg_cfg["lang"]
    train_path = data_cfg.get("train", None)
    dev_path = data_cfg.get("dev", None)
    test_path = data_cfg.get("test", None)

    if train_path is None and dev_path is None and test_path is None:
        raise ValueError("Please specify at least one data source path.")

    # build tokenizer
    logger.info("Building tokenizer...")
    tokenizer = build_tokenizer(data_cfg)

    dataset_type = data_cfg.get("dataset_type", "plain")
    dataset_cfg = data_cfg.get("dataset_cfg", {})

    # train data
    train_data = None
    if "train" in datasets and train_path is not None:
        logger.info("Loading train set...")
        train_data = build_dataset(
            dataset_type=dataset_type,
            path=train_path,
            src_lang=src_lang,
            trg_lang=trg_lang,
            split="train",
            tokenizer=tokenizer,
            random_subset=data_cfg.get("random_train_subset", -1),
            **dataset_cfg,
        )

    # build vocab
    logger.info("Building vocabulary...")
    src_vocab, trg_vocab = build_vocab(data_cfg, dataset=train_data)

    # set vocab to tokenizer
    tokenizer[src_lang].set_vocab(src_vocab._itos)  # pylint: disable=protected-access
    tokenizer[trg_lang].set_vocab(trg_vocab._itos)  # pylint: disable=protected-access

    # encoding func
    sequence_encoder = {
        src_lang: partial(src_vocab.sentences_to_ids, bos=False, eos=True),
        trg_lang: partial(trg_vocab.sentences_to_ids, bos=True, eos=True),
    }
    if train_data is not None:
        train_data.sequence_encoder = sequence_encoder

    # dev data
    dev_data = None
    if "dev" in datasets and dev_path is not None:
        logger.info("Loading dev set...")
        dev_data = build_dataset(
            dataset_type=dataset_type,
            path=dev_path,
            src_lang=src_lang,
            trg_lang=trg_lang,
            split="dev",
            tokenizer=tokenizer,
            sequence_encoder=sequence_encoder,
            random_subset=data_cfg.get("random_dev_subset", -1),
            **dataset_cfg,
        )

    # test data
    test_data = None
    if "test" in datasets and test_path is not None:
        logger.info("Loading test set...")
        test_data = build_dataset(
            dataset_type=dataset_type,
            path=test_path,
            src_lang=src_lang,
            trg_lang=trg_lang,
            split="test",
            tokenizer=tokenizer,
            sequence_encoder=sequence_encoder,
            random_subset=-1,  # no subsampling for test
            **dataset_cfg,
        )
    logger.info("Data loaded.")
    log_data_info(src_vocab, trg_vocab, train_data, dev_data, test_data)
    return src_vocab, trg_vocab, train_data, dev_data, test_data


def collate_fn(
    batch: List[Tuple],
    src_process: Callable,
    trg_process: Callable,
    pad_index: int = PAD_ID,
    device: torch.device = CPU_DEVICE,
    has_trg: bool = True,
    is_train: bool = True,
) -> Batch:
    """
    Custom collate function.
    See https://pytorch.org/docs/stable/data.html#dataloader-collate-fn for details.
    Note: you might need another collate_fn() if you switch to a different batch class.
    Please override the batch class here. (not in TrainManager)

    :param batch:
    :param src_process:
    :param trg_process:
    :param pad_index:
    :param device:
    :param has_trg:
    :param is_train:
    :return: joeynmt batch object
    """

    def _is_valid(s, t):
        # pylint: disable=no-else-return
        if has_trg:
            return s is not None and t is not None
        else:
            return s is not None

    batch = [(s, t) for s, t in batch if _is_valid(s, t)]
    src_list, trg_list = zip(*batch)
    assert len(batch) == len(src_list), (len(batch), len(src_list))
    assert all(s is not None for s in src_list), src_list
    src, src_length = src_process(src_list)

    if has_trg:
        assert all(t is not None for t in trg_list), trg_list
        assert trg_process is not None
        trg, trg_length = trg_process(trg_list)
    else:
        assert all(t is None for t in trg_list)
        trg, trg_length = None, None
    return Batch(
        src=torch.tensor(src).long(),
        src_length=torch.tensor(src_length).long(),
        trg=torch.tensor(trg).long() if trg else None,
        trg_length=torch.tensor(trg_length).long() if trg_length else None,
        device=device,
        pad_index=pad_index,
        has_trg=has_trg,
        is_train=is_train,
    )


def make_data_iter(
    dataset: Dataset,
    batch_size: int,
    batch_type: str = "sentence",
    seed: int = 42,
    shuffle: bool = False,
    num_workers: int = 0,
    pad_index: int = PAD_ID,
    device: torch.device = CPU_DEVICE,
) -> DataLoader:
    """
    Returns a torch DataLoader for a torch Dataset. (no bucketing)

    :param dataset: torch dataset containing src and optionally trg
    :param batch_size: size of the batches the iterator prepares
    :param batch_type: measure batch size by sentence count or by token count
    :param seed: random seed for shuffling
    :param shuffle: whether to shuffle the data before each epoch
        (for testing, no effect even if set to True)
    :param num_workers: number of cpus for multiprocessing
    :param pad_index:
    :param device:
    :return: torch DataLoader
    """
    assert isinstance(dataset, Dataset), dataset
    # sampler
    sampler: Sampler[int]  # (type annotation)
    if shuffle and dataset.split == "train":
        generator = torch.Generator()
        generator.manual_seed(seed)
        sampler = RandomSampler(dataset, generator=generator)
    else:
        sampler = SequentialSampler(dataset)

    # batch generator
    if batch_type == "sentence":
        batch_sampler = SentenceBatchSampler(sampler,
                                             batch_size=batch_size,
                                             drop_last=False)
    elif batch_type == "token":
        batch_sampler = TokenBatchSampler(sampler,
                                          batch_size=batch_size,
                                          drop_last=False)

    assert dataset.sequence_encoder[dataset.src_lang] is not None
    if dataset.has_trg:
        assert dataset.sequence_encoder[dataset.trg_lang] is not None
    else:
        dataset.sequence_encoder[dataset.trg_lang] = None

    # data iterator
    return DataLoader(
        dataset,
        batch_sampler=batch_sampler,
        collate_fn=partial(
            collate_fn,
            src_process=dataset.sequence_encoder[dataset.src_lang],
            trg_process=dataset.sequence_encoder[dataset.trg_lang],
            pad_index=pad_index,
            device=device,
            has_trg=dataset.has_trg,
            is_train=dataset.split == "train",
        ),
        num_workers=num_workers,
    )


class SentenceBatchSampler(BatchSampler):
    """
    Wraps another sampler to yield a mini-batch of indices based on num of instances.
    An instance longer than dataset.max_len will be filtered out.

    :param sampler: Base sampler. Can be any iterable object
    :param batch_size: Size of mini-batch.
    :param drop_last: If `True`, the sampler will drop the last batch if its size
        would be less than `batch_size`
    """

    def __init__(self, sampler: Sampler, batch_size: int, drop_last: bool):
        super().__init__(sampler, batch_size, drop_last)

    def __iter__(self):
        batch = []
        d = self.sampler.data_source
        for idx in self.sampler:
            src, trg = d[idx]  # pylint: disable=unused-variable
            if src is not None:  # otherwise drop instance
                batch.append(idx)
                if len(batch) >= self.batch_size:
                    yield batch
                    batch = []
        if len(batch) > 0 and not self.drop_last:
            yield batch


class TokenBatchSampler(BatchSampler):
    """
    Wraps another sampler to yield a mini-batch of indices based on num of tokens
    (incl. padding). An instance longer than dataset.max_len or shorter than
    dataset.min_len will be filtered out.
    * no bucketing implemented

    :param sampler: Base sampler. Can be any iterable object
    :param batch_size: Size of mini-batch.
    :param drop_last: If `True`, the sampler will drop the last batch if
            its size would be less than `batch_size`
    """

    def __init__(self, sampler: Sampler, batch_size: int, drop_last: bool):
        super().__init__(sampler, batch_size, drop_last)

    def __iter__(self):
        batch = []
        max_tokens = 0
        d = self.sampler.data_source
        for idx in self.sampler:
            src, trg = d[idx]  # call __getitem__()
            if src is not None:  # otherwise drop instance
                src_len = 0 if src is None else len(src)
                trg_len = 0 if trg is None else len(trg)
                n_tokens = 0 if src_len == 0 else max(src_len + 1, trg_len + 2)
                batch.append(idx)
                if n_tokens > max_tokens:
                    max_tokens = n_tokens
                if max_tokens * len(batch) >= self.batch_size:
                    yield batch
                    batch = []
                    max_tokens = 0
        if len(batch) > 0 and not self.drop_last:
            yield batch

    def __len__(self):
        raise NotImplementedError



from test.unit.test_helpers import TensorTestCase

import torch

from joeynmt.embeddings import Embeddings


class TestEmbeddings(TensorTestCase):

    def setUp(self):
        self.emb_size = 10
        self.vocab_size = 11
        self.pad_idx = 1
        seed = 42
        torch.manual_seed(seed)

    def test_size(self):
        emb = Embeddings(
            embedding_dim=self.emb_size,
            vocab_size=self.vocab_size,
            padding_idx=self.pad_idx,
        )
        self.assertEqual(emb.lut.weight.shape,
                         torch.Size([self.vocab_size, self.emb_size]))

    def test_pad_zeros(self):
        emb = Embeddings(
            embedding_dim=self.emb_size,
            vocab_size=self.vocab_size,
            padding_idx=self.pad_idx,
        )
        # pad embedding should be zeros
        self.assertTensorEqual(emb.lut.weight[self.pad_idx],
                               torch.zeros([self.emb_size]))

    def test_freeze(self):
        encoder = Embeddings(
            embedding_dim=self.emb_size,
            vocab_size=self.vocab_size,
            padding_idx=self.pad_idx,
            freeze=True,
        )
        for _, p in encoder.named_parameters():
            self.assertFalse(p.requires_grad)

    def test_forward(self):
        # fix the embedding weights
        weights = self._get_random_embedding_weights()
        emb = Embeddings(
            embedding_dim=self.emb_size,
            vocab_size=self.vocab_size,
            padding_idx=self.pad_idx,
        )
        emb.lut.weight.data = weights  # fill embeddings
        indices = torch.Tensor([0, 1, self.pad_idx, 9]).long()
        embedded = emb.forward(x=indices)
        # embedding operation is just slicing from weights matrix
        self.assertTensorEqual(embedded,
                               torch.index_select(input=weights, index=indices, dim=0))
        # after embedding, representations for PAD should still be zero
        self.assertTensorEqual(embedded[2], torch.zeros([self.emb_size]))

    def test_scale(self):
        # fix the embedding weights
        weights = self._get_random_embedding_weights()
        emb = Embeddings(
            embedding_dim=self.emb_size,
            vocab_size=self.vocab_size,
            padding_idx=self.pad_idx,
            scale=True,
        )
        emb.lut.weight.data = weights
        indices = torch.Tensor([0, 1, self.pad_idx, 9]).long()
        embedded = emb.forward(x=indices)
        # now scaled
        self.assertTensorNotEqual(
            torch.index_select(input=weights, index=indices, dim=0), embedded)
        self.assertTensorEqual(
            torch.index_select(input=weights, index=indices, dim=0) *
            (self.emb_size**0.5),
            embedded,
        )

    def _get_random_embedding_weights(self):
        weights = torch.rand([self.vocab_size, self.emb_size])
        weights[self.pad_idx] = torch.zeros([self.emb_size])
        return weights



# coding: utf-8
"""
Collection of builder functions
"""
import logging
from functools import partial
from typing import Callable, Generator, Optional

import torch
from torch import nn
from torch.optim import Optimizer
from torch.optim.lr_scheduler import (
    ExponentialLR,
    ReduceLROnPlateau,
    StepLR,
    _LRScheduler,
)

from joeynmt.helpers import ConfigurationError

logger = logging.getLogger(__name__)


def build_gradient_clipper(config: dict) -> Optional[Callable]:
    """
    Define the function for gradient clipping as specified in configuration.
    If not specified, returns None.

    Current options:
        - "clip_grad_val": clip the gradients if they exceed this value,
            see `torch.nn.utils.clip_grad_value_`
        - "clip_grad_norm": clip the gradients if their norm exceeds this value,
            see `torch.nn.utils.clip_grad_norm_`

    :param config: dictionary with training configurations
    :return: clipping function (in-place) or None if no gradient clipping
    """
    if "clip_grad_val" in config.keys() and "clip_grad_norm" in config.keys():
        raise ConfigurationError(
            "You can only specify either clip_grad_val or clip_grad_norm.")

    clip_grad_fun = None
    if "clip_grad_val" in config.keys():
        clip_grad_fun = partial(nn.utils.clip_grad_value_,
                                clip_value=config["clip_grad_val"])
    elif "clip_grad_norm" in config.keys():
        clip_grad_fun = partial(nn.utils.clip_grad_norm_,
                                max_norm=config["clip_grad_norm"])
    return clip_grad_fun


def build_optimizer(config: dict, parameters: Generator) -> Optimizer:
    """
    Create an optimizer for the given parameters as specified in config.

    Except for the weight decay and initial learning rate,
    default optimizer settings are used.

    Currently supported configuration settings for "optimizer":
        - "sgd" (default): see `torch.optim.SGD`
        - "adam": see `torch.optim.adam`
        - "adagrad": see `torch.optim.adagrad`
        - "adadelta": see `torch.optim.adadelta`
        - "rmsprop": see `torch.optim.RMSprop`

    The initial learning rate is set according to "learning_rate" in the config.
    The weight decay is set according to "weight_decay" in the config.
    If they are not specified, the initial learning rate is set to 3.0e-4, the
    weight decay to 0.

    Note that the scheduler state is saved in the checkpoint, so if you load
    a model for further training you have to use the same type of scheduler.

    :param config: configuration dictionary
    :param parameters:
    :return: optimizer
    """
    optimizer_name = config.get("optimizer", "sgd").lower()

    kwargs = {
        "lr": config.get("learning_rate", 3.0e-4),
        "weight_decay": config.get("weight_decay", 0),
    }

    if optimizer_name == "adam":
        kwargs["betas"] = config.get("adam_betas", (0.9, 0.999))
        optimizer = torch.optim.Adam(parameters, **kwargs)
    elif optimizer_name == "adagrad":
        optimizer = torch.optim.Adagrad(parameters, **kwargs)
    elif optimizer_name == "adadelta":
        optimizer = torch.optim.Adadelta(parameters, **kwargs)
    elif optimizer_name == "rmsprop":
        optimizer = torch.optim.RMSprop(parameters, **kwargs)
    elif optimizer_name == "sgd":
        # default
        kwargs["momentum"] = config.get("momentum", 0.0)
        optimizer = torch.optim.SGD(parameters, **kwargs)
    else:
        raise ConfigurationError("Invalid optimizer. Valid options: 'adam', "
                                 "'adagrad', 'adadelta', 'rmsprop', 'sgd'.")

    logger.info(
        "%s(%s)",
        optimizer.__class__.__name__,
        ", ".join([f"{k}={v}" for k, v in kwargs.items()]),
    )
    return optimizer


def build_scheduler(
    config: dict,
    optimizer: Optimizer,
    scheduler_mode: str,
    hidden_size: int = 0,
) -> (Optional[_LRScheduler], Optional[str]):
    """
    Create a learning rate scheduler if specified in config and determine when a
    scheduler step should be executed.

    Current options:
        - "plateau": see `torch.optim.lr_scheduler.ReduceLROnPlateau`
        - "decaying": see `torch.optim.lr_scheduler.StepLR`
        - "exponential": see `torch.optim.lr_scheduler.ExponentialLR`
        - "noam": see `joeynmt.builders.NoamScheduler`
        - "warmupexponentialdecay": see
          `joeynmt.builders.WarmupExponentialDecayScheduler`
        - "warmupinversesquareroot": see
          `joeynmt.builders.WarmupInverseSquareRootScheduler`

    If no scheduler is specified, returns (None, None) which will result in a constant
    learning rate.

    :param config: training configuration
    :param optimizer: optimizer for the scheduler, determines the set of parameters
        which the scheduler sets the learning rate for
    :param scheduler_mode: "min" or "max", depending on whether the validation score
        should be minimized or maximized. Only relevant for "plateau".
    :param hidden_size: encoder hidden size (required for NoamScheduler)
    :return:
        - scheduler: scheduler object,
        - scheduler_step_at: either "validation", "epoch", "step" or "none"
    """
    scheduler, scheduler_step_at = None, None
    if "scheduling" in config.keys() and config["scheduling"]:
        scheduler_name = config["scheduling"].lower()
        kwargs = {}
        if scheduler_name == "plateau":
            # learning rate scheduler
            kwargs = {
                "mode": scheduler_mode,
                "verbose": False,
                "threshold_mode": "abs",
                "eps": 0.0,
                "factor": config.get("decrease_factor", 0.1),
                "patience": config.get("patience", 10),
            }
            scheduler = ReduceLROnPlateau(optimizer=optimizer, **kwargs)
            # scheduler step is executed after every validation
            scheduler_step_at = "validation"
        elif scheduler_name == "decaying":
            kwargs = {"step_size": config.get("decaying_step_size", 1)}
            scheduler = StepLR(optimizer=optimizer, **kwargs)
            # scheduler step is executed after every epoch
            scheduler_step_at = "epoch"
        elif scheduler_name == "exponential":
            kwargs = {"gamma": config.get("decrease_factor", 0.99)}
            scheduler = ExponentialLR(optimizer=optimizer, **kwargs)
            # scheduler step is executed after every epoch
            scheduler_step_at = "epoch"
        elif scheduler_name == "noam":
            scheduler = NoamScheduler(
                optimizer=optimizer,
                hidden_size=hidden_size,
                factor=config.get("learning_rate_factor", 1),
                warmup=config.get("learning_rate_warmup", 4000),
            )
            scheduler_step_at = "step"
        elif scheduler_name == "warmupexponentialdecay":
            scheduler = WarmupExponentialDecayScheduler(
                min_rate=config.get("learning_rate_min", 1.0e-5),
                decay_rate=config.get("learning_rate_decay", 0.1),
                warmup=config.get("learning_rate_warmup", 4000),
                peak_rate=config.get("learning_rate_peak", 1.0e-3),
                decay_length=config.get("learning_rate_decay_length", 10000),
            )
            scheduler_step_at = "step"
        elif scheduler_name == "warmupinversesquareroot":
            lr = config.get("learning_rate", 1.0e-3)
            peak_rate = config.get("learning_rate_peak", lr)
            scheduler = WarmupInverseSquareRootScheduler(
                optimizer=optimizer,
                peak_rate=peak_rate,
                min_rate=config.get("learning_rate_min", 1.0e-5),
                warmup=config.get("learning_rate_warmup", 10000),
            )
            scheduler_step_at = "step"

    if scheduler is None:
        scheduler_step_at = "none"
    else:
        assert scheduler_step_at in {"validation", "epoch", "step", "none"}

    # print log
    if scheduler_name in [
            "noam",
            "warmupexponentialdecay",
            "warmupinversesquareroot",
    ]:
        logger.info(scheduler)
    else:
        logger.info(
            "%s(%s)",
            scheduler.__class__.__name__,
            ", ".join([f"{k}={v}" for k, v in kwargs.items()]),
        )
    return scheduler, scheduler_step_at


class BaseScheduler:
    """Base LR Scheduler
    decay at "step"
    """

    def __init__(self, optimizer: torch.optim.Optimizer):
        """
        :param optimizer:
        """
        self.optimizer = optimizer
        self._step = 0
        self._rate = 0
        self._state_dict = {"step": self._step, "rate": self._rate}

    def state_dict(self):
        """Returns dictionary of values necessary to reconstruct scheduler"""
        self._state_dict["step"] = self._step
        self._state_dict["rate"] = self._rate
        return self._state_dict

    def load_state_dict(self, state_dict):
        """Given a state_dict, this function loads scheduler's state"""
        self._step = state_dict["step"]
        self._rate = state_dict["rate"]

    def step(self, step):
        """Update parameters and rate"""
        self._step = step + 1  # sync with trainer.stats.steps
        rate = self._compute_rate()
        for p in self.optimizer.param_groups:
            p["lr"] = rate
        self._rate = rate

    def _compute_rate(self):
        raise NotImplementedError


class NoamScheduler(BaseScheduler):
    """
    The Noam learning rate scheduler used in "Attention is all you need"
    See Eq. 3 in https://arxiv.org/abs/1706.03762
    """

    def __init__(
        self,
        hidden_size: int,
        optimizer: torch.optim.Optimizer,
        factor: float = 1.0,
        warmup: int = 4000,
    ):
        """
        Warm-up, followed by learning rate decay.

        :param hidden_size:
        :param optimizer:
        :param factor: decay factor
        :param warmup: number of warmup steps
        """
        super().__init__(optimizer)
        self.warmup = warmup
        self.factor = factor
        self.hidden_size = hidden_size

    def _compute_rate(self):
        """Implement `lrate` above"""
        step = self._step
        upper_bound = min(step**(-0.5), step * self.warmup**(-1.5))
        return self.factor * (self.hidden_size**(-0.5) * upper_bound)

    def state_dict(self):
        """Returns dictionary of values necessary to reconstruct scheduler"""
        super().state_dict()
        self._state_dict["warmup"] = self.warmup
        self._state_dict["factor"] = self.factor
        self._state_dict["hidden_size"] = self.hidden_size
        return self._state_dict

    def load_state_dict(self, state_dict):
        """Given a state_dict, this function loads scheduler's state"""
        super().load_state_dict(state_dict)
        self.warmup = state_dict["warmup"]
        self.factor = state_dict["factor"]
        self.hidden_size = state_dict["hidden_size"]

    def __repr__(self):
        return (f"{self.__class__.__name__}(warmup={self.warmup}, "
                f"factor={self.factor}, hidden_size={self.hidden_size})")


class WarmupExponentialDecayScheduler(BaseScheduler):
    """
    A learning rate scheduler similar to Noam, but modified:
    Keep the warm up period but make it so that the decay rate can be tuneable.
    The decay is exponential up to a given minimum rate.
    """

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        peak_rate: float = 1.0e-3,
        decay_length: int = 10000,
        warmup: int = 4000,
        decay_rate: float = 0.5,
        min_rate: float = 1.0e-5,
    ):
        """
        Warm-up, followed by exponential learning rate decay.

        :param peak_rate: maximum learning rate at peak after warmup
        :param optimizer:
        :param decay_length: decay length after warmup
        :param decay_rate: decay rate after warmup
        :param warmup: number of warmup steps
        :param min_rate: minimum learning rate
        """
        super().__init__(optimizer)
        self.warmup = warmup
        self.decay_length = decay_length
        self.peak_rate = peak_rate
        self.decay_rate = decay_rate
        self.min_rate = min_rate

    def _compute_rate(self):
        """Implement `lrate` above"""
        step = self._step
        warmup = self.warmup

        if step < warmup:
            rate = step * self.peak_rate / warmup
        else:
            exponent = (step - warmup) / self.decay_length
            rate = self.peak_rate * (self.decay_rate**exponent)
        return max(rate, self.min_rate)

    def state_dict(self):
        """Returns dictionary of values necessary to reconstruct scheduler"""
        super().state_dict()
        self._state_dict["warmup"] = self.warmup
        self._state_dict["decay_length"] = self.decay_length
        self._state_dict["peak_rate"] = self.peak_rate
        self._state_dict["decay_rate"] = self.decay_rate
        self._state_dict["min_rate"] = self.min_rate
        return self._state_dict

    def load_state_dict(self, state_dict):
        """Given a state_dict, this function loads scheduler's state"""
        super().load_state_dict(state_dict)
        self.warmup = state_dict["warmup"]
        self.decay_length = state_dict["decay_length"]
        self.peak_rate = state_dict["peak_rate"]
        self.decay_rate = state_dict["decay_rate"]
        self.min_rate = state_dict["min_rate"]

    def __repr__(self):
        return (f"{self.__class__.__name__}(warmup={self.warmup}, "
                f"decay_length={self.decay_length}, "
                f"decay_rate={self.decay_rate}, "
                f"peak_rate={self.peak_rate}, "
                f"min_rate={self.min_rate})")


class WarmupInverseSquareRootScheduler(BaseScheduler):
    """
    Decay the LR based on the inverse square root of the update number.
    In the warmup phase, we linearly increase the learning rate.
    After warmup, we decrease the learning rate as follows:
    ```
    decay_factor = peak_rate * sqrt(warmup) # constant value
    lr = decay_factor / sqrt(step)
    ```
    cf.) https://github.com/pytorch/fairseq/blob/main/fairseq/optim/lr_scheduler/inverse_square_root_schedule.py
    """  # noqa

    def __init__(
        self,
        optimizer: torch.optim.Optimizer,
        peak_rate: float = 1.0e-3,
        warmup: int = 10000,
        min_rate: float = 1.0e-5,
    ):
        """
        Warm-up, followed by inverse square root learning rate decay.
        :param optimizer:
        :param peak_rate: maximum learning rate at peak after warmup
        :param warmup: number of warmup steps
        :param min_rate: minimum learning rate
        """
        super().__init__(optimizer)
        self.warmup = warmup
        self.min_rate = min_rate
        self.peak_rate = peak_rate
        self.decay_rate = peak_rate * (warmup**0.5)  # constant value

    def _compute_rate(self):
        """Implement `lrate` above"""
        step = self._step
        warmup = self.warmup

        if step < warmup:
            # linear warmup
            rate = step * self.peak_rate / warmup
        else:
            # decay prop. to the inverse square root of the update number
            rate = self.decay_rate * (step**-0.5)
        return max(rate, self.min_rate)

    def state_dict(self):
        """Returns dictionary of values necessary to reconstruct scheduler"""
        super().state_dict()
        self._state_dict["warmup"] = self.warmup
        self._state_dict["peak_rate"] = self.peak_rate
        self._state_dict["decay_rate"] = self.decay_rate
        self._state_dict["min_rate"] = self.min_rate
        return self._state_dict

    def load_state_dict(self, state_dict):
        """Given a state_dict, this function loads scheduler's state"""
        super().load_state_dict(state_dict)
        self.warmup = state_dict["warmup"]
        self.decay_rate = state_dict["decay_rate"]
        self.peak_rate = state_dict["peak_rate"]
        self.min_rate = state_dict["min_rate"]

    def __repr__(self):
        return (f"{self.__class__.__name__}(warmup={self.warmup}, "
                f"decay_rate={self.decay_rate:.6f}, peak_rate={self.peak_rate}, "
                f"min_rate={self.min_rate})")



import unittest

import torch
from torch import Tensor


class TensorTestCase(unittest.TestCase):

    def assertTensorNotEqual(self, expected: Tensor, actual: Tensor):
        equal = torch.equal(expected, actual)
        if equal:
            self.fail(f"Tensors did match but weren't supposed to: "
                      f"expected {expected}, actual {actual}.")

    def assertTensorEqual(self, expected: Tensor, actual: Tensor):
        equal = torch.equal(expected, actual)
        if not equal:
            self.fail(f"Tensors didn't match but were supposed to "
                      f"{expected} vs {actual}")

    def assertTensorAlmostEqual(self, expected: Tensor, actual: Tensor):
        diff = torch.all(torch.lt(torch.abs(torch.add(expected, -actual)), 1e-4))
        if not diff:
            self.fail(f"Tensors didn't match but were supposed to "
                      f"{expected} vs {actual}")



# coding: utf-8
"""
Implementation of a mini-batch.
"""
import logging
from typing import List, Optional

import numpy as np
import torch
from torch import Tensor

from joeynmt.constants import PAD_ID

logger = logging.getLogger(__name__)


class Batch:
    """
    Object for holding a batch of data with mask during training.
    Input is yielded from `collate_fn()` called by torch.data.utils.DataLoader.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        src: Tensor,
        src_length: Tensor,
        trg: Optional[Tensor],
        trg_length: Optional[Tensor],
        device: torch.device,
        pad_index: int = PAD_ID,
        has_trg: bool = True,
        is_train: bool = True,
    ):
        """
        Creates a new joey batch. This batch supports attributes with src and trg
        length, masks, number of non-padded tokens in trg. Furthermore, it can be
        sorted by src length.

        :param src:
        :param src_length:
        :param trg:
        :param trg_length:
        :param device:
        :param pad_index: *must be the same for both src and trg
        :param is_train: *can be used for online data augmentation, subsampling etc.
        """
        self.src: Tensor = src
        self.src_length: Tensor = src_length
        self.src_mask: Tensor = (self.src != pad_index).unsqueeze(1)
        self.trg_input: Optional[Tensor] = None
        self.trg: Optional[Tensor] = None
        self.trg_mask: Optional[Tensor] = None
        self.trg_length: Optional[Tensor] = None

        self.nseqs: int = self.src.size(0)
        self.ntokens: Optional[int] = None
        self.has_trg: bool = has_trg
        self.is_train: bool = is_train

        if self.has_trg:
            assert trg is not None and trg_length is not None
            # trg_input is used for teacher forcing, last one is cut off
            self.trg_input: Tensor = trg[:, :-1]  # shape (batch_size, seq_length)
            self.trg_length: Tensor = trg_length - 1
            # trg is used for loss computation, shifted by one since BOS
            self.trg: Tensor = trg[:, 1:]  # shape (batch_size, seq_length)
            # we exclude the padded areas (and blank areas) from the loss computation
            self.trg_mask: Tensor = (self.trg != pad_index).unsqueeze(1)
            self.ntokens: int = (self.trg != pad_index).data.sum().item()

        if device.type == "cuda":
            self._make_cuda(device)

        # a batch has to contain more than one src sentence
        assert self.nseqs > 0, self.nseqs

    def _make_cuda(self, device: torch.device) -> None:
        """Move the batch to GPU"""
        self.src = self.src.to(device)
        self.src_length = self.src_length.to(device)
        self.src_mask = self.src_mask.to(device)

        if self.has_trg:
            self.trg_input = self.trg_input.to(device)
            self.trg = self.trg.to(device)
            self.trg_mask = self.trg_mask.to(device)

    def normalize(
        self,
        tensor: Tensor,
        normalization: str = "none",
        n_gpu: int = 1,
        n_accumulation: int = 1,
    ) -> Tensor:
        """
        Normalizes batch tensor (i.e. loss). Takes sum over multiple gpus, divides by
        nseqs or ntokens, divide by n_gpu, then divide by n_accumulation.

        :param tensor: (Tensor) tensor to normalize, i.e. batch loss
        :param normalization: (str) one of {`batch`, `tokens`, `none`}
        :param n_gpu: (int) the number of gpus
        :param n_accumulation: (int) the number of gradient accumulation
        :return: normalized tensor
        """
        if n_gpu > 1:
            tensor = tensor.sum()

        if normalization == "sum":  # pylint: disable=no-else-return
            return tensor
        elif normalization == "batch":
            normalizer = self.nseqs
        elif normalization == "tokens":
            normalizer = self.ntokens
        elif normalization == "none":
            normalizer = 1

        norm_tensor = tensor / normalizer

        if n_gpu > 1:
            norm_tensor = norm_tensor / n_gpu

        if n_accumulation > 1:
            norm_tensor = norm_tensor / n_accumulation
        return norm_tensor

    def sort_by_src_length(self) -> List[int]:
        """
        Sort by src length (descending) and return index to revert sort

        :return: list of indices
        """
        _, perm_index = self.src_length.sort(0, descending=True)
        rev_index = [0] * perm_index.size(0)
        for new_pos, old_pos in enumerate(perm_index.cpu().numpy()):
            rev_index[old_pos] = new_pos

        sorted_src_length = self.src_length[perm_index]
        sorted_src = self.src[perm_index]
        sorted_src_mask = self.src_mask[perm_index]
        if self.has_trg:
            sorted_trg_input = self.trg_input[perm_index]
            sorted_trg_length = self.trg_length[perm_index]
            sorted_trg_mask = self.trg_mask[perm_index]
            sorted_trg = self.trg[perm_index]

        self.src = sorted_src
        self.src_length = sorted_src_length
        self.src_mask = sorted_src_mask

        if self.has_trg:
            self.trg_input = sorted_trg_input
            self.trg_mask = sorted_trg_mask
            self.trg_length = sorted_trg_length
            self.trg = sorted_trg

        assert max(rev_index) < len(rev_index), rev_index
        return rev_index

    def score(self, log_probs: Tensor) -> np.ndarray:
        """Look up the score of the trg token (ground truth) in the batch"""
        scores = []
        for i in range(self.nseqs):
            scores.append(
                np.array([
                    log_probs[i, j, ind].item() for j, ind in enumerate(self.trg[i])
                    if ind != PAD_ID
                ]))
        # Note: each element in `scores` list can have different lengths.
        return np.array(scores, dtype=object)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(nseqs={self.nseqs}, ntokens={self.ntokens}, "
            f"has_trg={self.has_trg}, is_train={self.is_train})")



# coding: utf-8
from setuptools import find_packages, setup

with open("requirements.txt", encoding="utf-8") as req_fp:
    install_requires = req_fp.readlines()

setup(
    name='joeynmt',
    version='2.0.0',
    description='Minimalist NMT for educational purposes',
    author='Jasmijn Bastings and Julia Kreutzer',
    url='https://github.com/joeynmt/joeynmt',
    license='Apache License',
    install_requires=install_requires,
    packages=find_packages(exclude=[]),
    python_requires='>=3.9',
    project_urls={
        'Documentation': 'http://joeynmt.readthedocs.io/en/latest/',
        'Source': 'https://github.com/joeynmt/joeynmt',
        'Tracker': 'https://github.com/joeynmt/joeynmt/issues',
    },
    entry_points={
        'console_scripts': [],
    }
)



# coding: utf-8
"""
Evaluation metrics
"""
import logging
from inspect import getfullargspec
from typing import List

from sacrebleu.metrics import BLEU, CHRF

logger = logging.getLogger(__name__)


def chrf(hypotheses: List[str], references: List[str], **sacrebleu_cfg) -> float:
    """
    Character F-score from sacrebleu
    cf. https://github.com/mjpost/sacrebleu/blob/master/sacrebleu/metrics/chrf.py

    :param hypotheses: list of hypotheses (strings)
    :param references: list of references (strings)
    :return: character f-score (0 <= chf <= 1)
             see Breaking Change in sacrebleu v2.0
    """
    kwargs = {}
    if sacrebleu_cfg:
        valid_keys = getfullargspec(CHRF).args
        for k, v in sacrebleu_cfg.items():
            if k in valid_keys:
                kwargs[k] = v

    metric = CHRF(**kwargs)
    score = metric.corpus_score(hypotheses=hypotheses, references=[references]).score

    # log sacrebleu signature
    logger.info(metric.get_signature())
    return score / 100


def bleu(hypotheses: List[str], references: List[str], **sacrebleu_cfg) -> float:
    """
    Raw corpus BLEU from sacrebleu (without tokenization)
    cf. https://github.com/mjpost/sacrebleu/blob/master/sacrebleu/metrics/bleu.py

    :param hypotheses: list of hypotheses (strings)
    :param references: list of references (strings)
    :return: bleu score
    """
    kwargs = {}
    if sacrebleu_cfg:
        valid_keys = getfullargspec(BLEU).args
        for k, v in sacrebleu_cfg.items():
            if k in valid_keys:
                kwargs[k] = v

    metric = BLEU(**kwargs)
    score = metric.corpus_score(hypotheses=hypotheses, references=[references]).score

    # log sacrebleu signature
    logger.info(metric.get_signature())
    return score


def token_accuracy(hypotheses: List[List[str]], references: List[List[str]]) -> float:
    """
    Compute the accuracy of hypothesis tokens: correct tokens / all tokens
    Tokens are correct if they appear in the same position in the reference.
    We lookup the references before one-hot-encoding, that is, UNK generation in
    hypotheses is always evaluated as incorrect.

    :param hypotheses: list of tokenized hypotheses (List[List[str]])
    :param references: list of tokenized references (List[List[str]])
    :return: token accuracy (float)
    """
    correct_tokens = 0
    all_tokens = 0
    assert len(hypotheses) == len(references)
    for hyp, ref in zip(hypotheses, references):
        all_tokens += len(hyp)
        for h_i, r_i in zip(hyp, ref):
            # min(len(h), len(r)) tokens considered
            if h_i == r_i:
                correct_tokens += 1
    return (correct_tokens / all_tokens) * 100 if all_tokens > 0 else 0.0


def sequence_accuracy(hypotheses: List[str], references: List[str]) -> float:
    """
    Compute the accuracy of hypothesis tokens: correct tokens / all tokens
    Tokens are correct if they appear in the same position in the reference.
    We lookup the references before one-hot-encoding, that is, hypotheses with UNK
    are always evaluated as incorrect.

    :param hypotheses: list of hypotheses (strings)
    :param references: list of references (strings)
    :return:
    """
    assert len(hypotheses) == len(references)
    correct_sequences = sum(
        [1 for (hyp, ref) in zip(hypotheses, references) if hyp == ref])
    return (correct_sequences / len(hypotheses)) * 100 if hypotheses else 0.0



from test.unit.test_helpers import TensorTestCase

import torch

from joeynmt.loss import XentLoss


class TestXentLoss(TensorTestCase):

    def setUp(self):
        seed = 42
        torch.manual_seed(seed)

    def test_label_smoothing(self):
        pad_index = 0
        smoothing = 0.4
        criterion = XentLoss(pad_index=pad_index, smoothing=smoothing)

        # batch x seq_len x vocab_size: 3 x 2 x 5
        predict = torch.FloatTensor([
            [[0.1, 0.1, 0.6, 0.1, 0.1], [0.1, 0.1, 0.6, 0.1, 0.1]],
            [[0.1, 0.1, 0.6, 0.1, 0.1], [0.1, 0.1, 0.6, 0.1, 0.1]],
            [[0.1, 0.1, 0.6, 0.1, 0.1], [0.1, 0.1, 0.6, 0.1, 0.1]],
        ])

        # batch x seq_len: 3 x 2
        targets = torch.LongTensor([[2, 1], [2, 0], [1, 0]])

        # test the smoothing function
        # pylint: disable=protected-access
        smoothed_targets = criterion._smooth_targets(targets=targets.view(-1),
                                                     vocab_size=predict.size(-1))
        # pylint: enable=protected-access
        self.assertTensorAlmostEqual(
            smoothed_targets,
            torch.Tensor([
                [0.0000, 0.1333, 0.6000, 0.1333, 0.1333],
                [0.0000, 0.6000, 0.1333, 0.1333, 0.1333],
                [0.0000, 0.1333, 0.6000, 0.1333, 0.1333],
                [0.0000, 0.0000, 0.0000, 0.0000, 0.0000],
                [0.0000, 0.6000, 0.1333, 0.1333, 0.1333],
                [0.0000, 0.0000, 0.0000, 0.0000, 0.0000],
            ]),
        )
        assert torch.max(smoothed_targets) == 1 - smoothing

        # test the loss computation
        v = criterion(predict.log(), **{"trg": targets})
        self.assertTensorAlmostEqual(v, 2.1326)

    def test_no_label_smoothing(self):
        pad_index = 0
        smoothing = 0.0
        criterion = XentLoss(pad_index=pad_index, smoothing=smoothing)

        # batch x seq_len x vocab_size: 3 x 2 x 5
        predict = torch.FloatTensor([
            [[0.1, 0.1, 0.6, 0.1, 0.1], [0.1, 0.1, 0.6, 0.1, 0.1]],
            [[0.1, 0.1, 0.6, 0.1, 0.1], [0.1, 0.1, 0.6, 0.1, 0.1]],
            [[0.1, 0.1, 0.6, 0.1, 0.1], [0.1, 0.1, 0.6, 0.1, 0.1]],
        ])

        # batch x seq_len: 3 x 2
        targets = torch.LongTensor([[2, 1], [2, 0], [1, 0]])

        # test the smoothing function: should still be one-hot
        # pylint: disable=protected-access
        smoothed_targets = criterion._smooth_targets(targets=targets.view(-1),
                                                     vocab_size=predict.size(-1))
        # pylint: enable=protected-access

        assert torch.max(smoothed_targets) == 1
        assert torch.min(smoothed_targets) == 0

        self.assertTensorAlmostEqual(
            smoothed_targets,
            torch.Tensor([
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0, 0.0],
            ]),
        )

        v = criterion(predict.log(), **{"trg": targets})
        self.assertTensorAlmostEqual(v, 5.6268)



from test.unit.test_helpers import TensorTestCase

import numpy as np
import torch

from joeynmt.decoders import RecurrentDecoder, TransformerDecoder
from joeynmt.embeddings import Embeddings
from joeynmt.encoders import RecurrentEncoder
from joeynmt.model import Model
from joeynmt.search import beam_search, greedy
from joeynmt.vocabulary import Vocabulary


class TestSearch(TensorTestCase):
    # pylint: disable=too-many-instance-attributes
    def setUp(self):
        self.emb_size = 12
        self.num_layers = 3
        self.hidden_size = 12
        self.ff_size = 24
        self.num_heads = 4
        self.dropout = 0.0
        self.encoder_hidden_size = 3
        self.vocab = Vocabulary(tokens=["word"])
        self.vocab_size = len(self.vocab)  # = 5
        seed = 42
        torch.manual_seed(seed)
        # self.bos_index = 2
        self.pad_index = 1
        # self.eos_index = 3

        self.expected_transformer_ids = [[5, 5, 5], [5, 5, 5]]
        self.expected_transformer_scores = np.array([
            [-1.256322, -1.3881024, -1.4247599],
            [-1.23621, -1.384755, -1.4188296],
        ])

        self.expected_recurrent_ids = [[4, 0, 4], [4, 4, 4]]
        self.expected_recurrent_scores = np.array([
            [-1.1915066, -1.2217927, -1.244617],
            [-1.1754444, -1.2138686, -1.204663],
        ])


class TestSearchTransformer(TestSearch):

    def _build(self, batch_size):
        src_time_dim = 4
        vocab_size = 7

        emb = Embeddings(
            embedding_dim=self.emb_size,
            vocab_size=vocab_size,
            padding_idx=self.pad_index,
        )

        decoder = TransformerDecoder(
            num_layers=self.num_layers,
            num_heads=self.num_heads,
            hidden_size=self.hidden_size,
            ff_size=self.ff_size,
            dropout=self.dropout,
            emb_dropout=self.dropout,
            vocab_size=vocab_size,
            layer_norm="pre",
        )

        encoder_output = torch.rand(size=(batch_size, src_time_dim, self.hidden_size))

        for p in decoder.parameters():
            torch.nn.init.uniform_(p, -0.5, 0.5)

        src_mask = torch.ones(size=(batch_size, 1, src_time_dim)) == 1

        encoder_hidden = None  # unused

        model = Model(
            encoder=None,
            decoder=decoder,
            src_embed=emb,
            trg_embed=emb,
            src_vocab=self.vocab,
            trg_vocab=self.vocab,
        )
        return src_mask, model, encoder_output, encoder_hidden

    def test_transformer_greedy(self):
        batch_size = 2
        max_output_length = 3
        src_mask, model, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        output, scores, attention_scores = greedy(
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            return_prob="hyp",
        )
        # Transformer greedy doesn't return attention scores
        # `return_attention = False` by default
        self.assertIsNone(attention_scores)

        # outputs
        self.assertEqual(output.shape, (batch_size, max_output_length))  # batch x time
        np.testing.assert_equal(output, self.expected_transformer_ids)

        # scores
        self.assertEqual(scores.shape, (batch_size, max_output_length))  # batch x time
        np.testing.assert_allclose(scores, self.expected_transformer_scores, rtol=1e-5)

    def test_transformer_beam1(self):
        batch_size = 2
        beam_size = 1
        alpha = 0.0
        n_best = 1
        max_output_length = 3
        src_mask, model, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        beam_output, beam_scores, attention_scores = beam_search(
            beam_size=beam_size,
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            alpha=alpha,
            n_best=n_best,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            return_prob="hyp",
        )
        # Transformer beam doesn't return attention scores
        self.assertIsNone(attention_scores)

        # batch_size * n_best x hyp_len
        self.assertEqual(beam_output.shape, (batch_size * n_best, max_output_length))
        np.testing.assert_equal(beam_output, self.expected_transformer_ids)
        np.testing.assert_allclose(beam_scores, [[-4.174977], [-4.141973]], rtol=1e-5)

        # now compare to greedy, they should be the same for beam=1
        greedy_output, greedy_scores, _ = greedy(
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            return_prob="hyp",
        )
        np.testing.assert_equal(beam_output, greedy_output)
        np.testing.assert_allclose(
            greedy_scores,
            self.expected_transformer_scores,
            rtol=1e-5,
        )

    def test_transformer_beam7(self):
        batch_size = 2
        beam_size = 7
        n_best = 5
        alpha = 1.0
        max_output_length = 3
        src_mask, model, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        output, scores, attention_scores = beam_search(
            beam_size=beam_size,
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            alpha=alpha,
            n_best=n_best,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            return_prob="hyp",
        )
        # Transformer beam doesn't return attention scores
        self.assertIsNone(attention_scores)

        # batch_size * n_best x hyp_len (= time steps)
        self.assertEqual(output.shape, (batch_size * n_best, max_output_length))
        expected_output = [[5, 5, 5], [0, 5, 5], [0, 0, 5], [5, 5, 0], [5, 0, 5],
                           [5, 5, 5], [0, 5, 5], [5, 0, 5], [5, 5, 0], [0, 0, 5]]
        np.testing.assert_equal(output, expected_output)
        expected_scores = [
            [-3.13123298],
            [-3.29512906],
            [-3.43877649],
            [-3.44861484],
            [-3.45595121],
            [-3.10648012],
            [-3.30023503],
            [-3.43445206],
            [-3.43654943],
            [-3.47406816],
        ]
        np.testing.assert_allclose(scores, expected_scores, rtol=1e-5)

    def test_repetition_penalty_and_generate_unk(self):
        batch_size = 3
        max_output_length = 3
        src_mask, model, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        # no repetition penalty
        output, _, _ = greedy(
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            generate_unk=False,
        )
        expected_output = [[5, 6, 4], [5, 6, 4], [5, 6, 6]]
        np.testing.assert_equal(output, expected_output)
        np.testing.assert_equal(np.count_nonzero(output), 9)  # no unk token generated

        # trg repetition penalty
        output_trg_penalty, _, _ = greedy(
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            encoder_input=None,
            repetition_penalty=1.5,
            generate_unk=False,
        )
        expected_output_trg_penalty = [[5, 6, 4], [5, 6, 4], [5, 6, 4]]
        np.testing.assert_equal(output_trg_penalty, expected_output_trg_penalty)

        # src + trg repetition penalty
        # src_len = 4 (see self._build())
        src_tokens = torch.tensor([[4, 3, 1, 1], [5, 4, 3, 1], [5, 5, 6, 3]])
        src_mask = (src_tokens != 1).unsqueeze(1)
        output_src_penalty, _, attention = greedy(
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            encoder_input=src_tokens,
            repetition_penalty=1.5,
            generate_unk=False,
            return_attention=True,
        )
        expected_output_src_penalty = [[5, 6, 4], [6, 4, 6], [4, 5, 6]]
        np.testing.assert_equal(output_src_penalty, expected_output_src_penalty)

        # Transformer Greedy can return attention probs
        expected_attention = np.array([[[0.50552493, 0.49447513, 0.0, 0.0],
                                        [0.49927852, 0.50072145, 0.0, 0.0],
                                        [0.4978183, 0.5021817, 0.0, 0.0]],
                                       [[0.33627957, 0.3340565, 0.32966393, 0.0],
                                        [0.33195895, 0.34128872, 0.32675233, 0.0],
                                        [0.33139172, 0.33966494, 0.3289433, 0.0]],
                                       [[0.25297716, 0.24552101, 0.26504526, 0.2364566],
                                        [0.25382724, 0.24408112, 0.2637272, 0.23836446],
                                        [0.24418367, 0.2407347, 0.26917717, 0.24590445]]
                                       ])  # (batch_size, trg_len, src_len) = (3, 3, 4)
        np.testing.assert_allclose(attention, expected_attention, rtol=1e-5)

    def test_repetition_penalty_in_beam_search(self):
        batch_size = 2
        beam_size = 7
        n_best = 5
        alpha = 1.0
        max_output_length = 3
        src_mask, model, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        src_tokens = torch.tensor([[5, 5, 4], [5, 6, 6]])
        expected_output_with_penalty = [
            [6, 0, 5],
            [6, 0, 1],
            [6, 1, 0],
            [6, 0, 0],
            [0, 5, 6],
            [5, 1, 0],
            [1, 0, 5],
            [4, 0, 5],
            [0, 5, 1],
            [0, 5, 3],
        ]
        expected_scores_with_penalty = [
            [-3.97227573],
            [-4.04966927],
            [-4.13298225],
            [-4.1490984],
            [-4.20313644],
            [-4.25672293],
            [-4.28554916],
            [-4.31342602],
            [-4.3269372],
            [-4.34846115],
        ]
        output_with_penalty, scores_with_penalty, _ = beam_search(
            beam_size=beam_size,
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            alpha=alpha,
            n_best=n_best,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            encoder_input=src_tokens,
            repetition_penalty=1.5,
            return_prob="hyp",
        )
        np.testing.assert_equal(output_with_penalty, expected_output_with_penalty)
        np.testing.assert_allclose(scores_with_penalty,
                                   expected_scores_with_penalty,
                                   rtol=1e-5)

    def test_ngram_blocker(self):
        batch_size = 2
        max_output_length = 10
        no_repeat_ngram_size = 3
        src_mask, model, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        expected_output = [[5, 5, 5, 0, 5, 5, 0, 0, 5, 0],
                           [5, 5, 5, 0, 5, 5, 0, 0, 5, 0]]
        expected_scores = [[
            -1.256322, -1.3881024, -1.4247599, -1.7823244, -1.3601547, -1.4100374,
            -1.8323467, -1.4702327, -1.3825183, -1.8272306
        ],
                           [
                               -1.2362098, -1.384755, -1.4188296, -1.7987821,
                               -1.3441339, -1.4011306, -1.845073, -1.5097935,
                               -1.3674619, -1.8354796
                           ]]
        output, scores, _ = greedy(
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            encoder_input=None,
            return_prob="hyp",
            no_repeat_ngram_size=no_repeat_ngram_size,
        )
        np.testing.assert_equal(output, expected_output)
        np.testing.assert_allclose(scores, expected_scores, rtol=1e-5)

    def test_ngram_blocker_in_beam_search(self):
        batch_size = 2
        beam_size = 3
        n_best = 3
        alpha = 1.0
        max_output_length = 10
        no_repeat_ngram_size = 3
        src_mask, model, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        expected_output = [
            [5, 5, 5, 0, 0, 0, 0, 0, 0, 5],
            [5, 5, 5, 0, 0, 0, 0, 0, 5, 5],
            [5, 5, 5, 0, 0, 0, 0, 0, 0, 0],
            [5, 5, 5, 0, 0, 5, 5, 0, 0, 0],
            [5, 5, 5, 0, 0, 0, 0, 5, 5, 0],
            [5, 5, 5, 0, 0, 0, 0, 5, 5, 2],
        ]
        expected_scores = [[-5.882798], [-5.884224], [-5.889960], [-5.969066],
                           [-6.015323], [-6.062648]]
        output, scores, _ = beam_search(
            beam_size=beam_size,
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            alpha=alpha,
            n_best=n_best,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            encoder_input=None,
            return_prob="hyp",
            no_repeat_ngram_size=no_repeat_ngram_size,
        )
        np.testing.assert_equal(output, expected_output)
        np.testing.assert_allclose(scores, expected_scores, rtol=1e-5)


class TestSearchRecurrent(TestSearch):

    def _build(self, batch_size):
        src_time_dim = 4
        vocab_size = 7

        emb = Embeddings(
            embedding_dim=self.emb_size,
            vocab_size=vocab_size,
            padding_idx=self.pad_index,
        )

        encoder = RecurrentEncoder(
            emb_size=self.emb_size,
            num_layers=self.num_layers,
            hidden_size=self.encoder_hidden_size,
            bidirectional=True,
        )

        decoder = RecurrentDecoder(
            hidden_size=self.hidden_size,
            encoder=encoder,
            attention="bahdanau",
            emb_size=self.emb_size,
            vocab_size=self.vocab_size,
            num_layers=self.num_layers,
            init_hidden="bridge",
            input_feeding=True,
        )

        encoder_output = torch.rand(size=(batch_size, src_time_dim,
                                          encoder.output_size))

        for p in decoder.parameters():
            torch.nn.init.uniform_(p, -0.5, 0.5)

        src_mask = torch.ones(size=(batch_size, 1, src_time_dim)) == 1

        encoder_hidden = torch.rand(size=(batch_size, encoder.output_size))

        model = Model(
            encoder=encoder,
            decoder=decoder,
            src_embed=emb,
            trg_embed=emb,
            src_vocab=self.vocab,
            trg_vocab=self.vocab,
        )

        return src_mask, model, encoder_output, encoder_hidden

    def test_recurrent_greedy(self):
        batch_size = 2
        max_output_length = 3
        src_mask, model, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        output, scores, attention_scores = greedy(
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            return_prob="hyp",
        )
        self.assertEqual(output.shape, (batch_size, max_output_length))
        np.testing.assert_equal(output, self.expected_recurrent_ids)
        np.testing.assert_allclose(scores, self.expected_recurrent_scores, rtol=1e-5)

        expected_attention_scores = np.array(
            [[[0.22914883, 0.24638498, 0.21247596, 0.3119903],
              [0.22970565, 0.24540883, 0.21261126, 0.31227428],
              [0.22903332, 0.2459198, 0.2110187, 0.3140282]],
             [[0.252522, 0.29074305, 0.257121, 0.19961396],
              [0.2519883, 0.2895494, 0.25718424, 0.201278],
              [0.2523954, 0.28959078, 0.25769445, 0.2003194]]])
        np.testing.assert_array_almost_equal(attention_scores,
                                             expected_attention_scores)
        self.assertEqual(attention_scores.shape, (batch_size, max_output_length, 4))

    def test_recurrent_beam1(self):
        # beam=1 and greedy should return the same result
        batch_size = 2
        max_output_length = 3
        src_mask, model, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        greedy_output, greedy_scores, _ = greedy(
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            return_prob="hyp",
        )
        self.assertEqual(greedy_output.shape, (batch_size, max_output_length))
        np.testing.assert_equal(greedy_output, self.expected_recurrent_ids)
        np.testing.assert_allclose(
            greedy_scores,
            self.expected_recurrent_scores,
            rtol=1e-5,
        )

        beam_size = 1
        alpha = 0.0
        n_best = 1
        beam_output, beam_scores, _ = beam_search(
            beam_size=beam_size,
            src_mask=src_mask,
            n_best=n_best,
            max_output_length=max_output_length,
            model=model,
            alpha=alpha,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            return_prob="hyp",
        )
        np.testing.assert_array_equal(greedy_output, beam_output)
        np.testing.assert_allclose(
            beam_scores,
            self.expected_recurrent_scores.sum(axis=1, keepdims=True),
            rtol=1e-5,
        )

    def test_recurrent_beam7(self):
        batch_size = 2
        max_output_length = 3
        src_mask, model, encoder_output, encoder_hidden = self._build(
            batch_size=batch_size)

        beam_size = 7
        n_best = 5
        alpha = 1.0
        output, scores, _ = beam_search(
            beam_size=beam_size,
            src_mask=src_mask,
            max_output_length=max_output_length,
            model=model,
            alpha=alpha,
            n_best=n_best,
            encoder_output=encoder_output,
            encoder_hidden=encoder_hidden,
            return_prob="hyp",
        )

        self.assertEqual(output.shape, (batch_size * n_best, max_output_length))

        # output indices
        expected_output = [[4, 4, 4], [4, 4, 0], [4, 0, 4], [4, 0, 0], [0, 4, 4],
                           [4, 4, 4], [4, 4, 0], [4, 0, 4], [4, 0, 0], [0, 4, 4]]
        np.testing.assert_array_equal(output, expected_output)

        # log probabilities
        expected_scores = [
            [-2.71620679],
            [-2.72217512],
            [-2.74343705],
            [-2.76944518],
            [-2.86219954],
            [-2.69548202],
            [-2.72114182],
            [-2.76927805],
            [-2.82477784],
            [-2.87750268],
        ]
        np.testing.assert_allclose(scores, expected_scores, rtol=1e-5)
